from typing import Any
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AppGitRepository(BaseModel):
    """App Git Repository settings"""

    provider: str = Field(..., description="")
    url: str = Field(..., description="")


# class AppProviderConfig(BaseModel):
#     """App Provider Configuration"""
#     workspace_id: str = Field(..., description="")


class AppResourceDatabase(BaseModel):
    """App Resource Database Configuration"""

    database_name: str = Field(..., description="The name of database.")
    instance_name: str = Field(..., description="The name of database instance.")
    permission: str = Field(
        ...,
        description="Permission to grant on database. Supported permissions are: CAN_CONNECT_AND_CREATE.",
    )


class AppResourceExperiment(BaseModel):
    """App Resource Experiment Configuration"""

    experiment_id: str = Field(..., description="")
    permission: str = Field(..., description="")


class AppResourceGenieSpace(BaseModel):
    """App Resource Genie Space Configuration"""

    name: str = Field(..., description="The name of Genie Space.")
    permission: str = Field(..., description="")
    space_id: str = Field(..., description="The unique ID of Genie Space.")


class AppResourceJob(BaseModel):
    """App Resource Job Configuration"""

    id: str = Field(..., description="Id of the job to grant permission on.")
    permission: str = Field(
        ...,
        description="Permissions to grant on the Job. Supported permissions are: `CAN_MANAGE`, `IS_OWNER`, `CAN_MANAGE_RUN`, `CAN_VIEW`.",
    )


class AppResourceSecret(BaseModel):
    """App Resource Secret Configuration"""

    key: str = Field(..., description="Key of the secret to grant permission on.")
    permission: str = Field(
        ...,
        description="Permission to grant on the secret scope. For secrets, only one permission is allowed. Permission must be one of: `READ`, `WRITE`, `MANAGE`.",
    )
    scope: str = Field(..., description="Scope of the secret to grant permission on.")


class AppResourceServingEndpoint(BaseModel):
    """App Resource Serving Endpoint Configuration"""

    name: str = Field(
        ..., description="Name of the serving endpoint to grant permission on."
    )
    permission: str = Field(
        ...,
        description="Permission to grant on the serving endpoint. Supported permissions are: `CAN_MANAGE`, `CAN_QUERY`, `CAN_VIEW`.",
    )


class AppResourceSqlWarehouse(BaseModel):
    """App Resource SQL Warehouse Configuration"""

    id: str = Field(..., description="Id of the SQL warehouse to grant permission on.")
    permission: str = Field(
        ...,
        description="Permission to grant on the SQL warehouse. Supported permissions are: `CAN_MANAGE`, `CAN_USE`, `IS_OWNER`.",
    )


class AppResourceUcSecurable(BaseModel):
    """App Resource Unity Catalog Securable Configuration"""

    permission: str = Field(
        ...,
        description="Permissions to grant on UC securable, i.e. `READ_VOLUME`, `WRITE_VOLUME`.",
    )
    securable_full_name: str = Field(
        ...,
        description="the full name of UC securable, i.e. `my-catalog.my-schema.my-volume`.",
    )
    securable_type: str = Field(
        ..., description="the type of UC securable, i.e. `VOLUME`."
    )
    securable_kind: str = Field(None, description="")


class AppResource(BaseModel):
    """App Resource Configuration"""

    name: str = Field(..., description="The name of the resource.")
    # app: AppResourceApp = Field(None, description="")
    database: AppResourceDatabase = Field(None, description="")
    description: str = Field(
        None,
        description="The description of the resource. Exactly one of the following attributes must be provided.",
    )
    experiment: AppResourceExperiment = Field(None, description="")
    genie_space: AppResourceGenieSpace = Field(None, description="attribute")
    job: AppResourceJob = Field(None, description="attribute")
    secret: AppResourceSecret = Field(None, description="attribute")
    serving_endpoint: AppResourceServingEndpoint = Field(None, description="attribute")
    sql_warehouse: AppResourceSqlWarehouse = Field(None, description="attribute")
    uc_securable: AppResourceUcSecurable = Field(
        None,
        description="attribute (see the API docs for full list of supported UC objects)",
    )


class App(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks App

    Examples
    --------
    ```py
    import io

    from laktory import models

    # Define App
    job_yaml = '''
    name: stocks-dash
    description: A dashboard app for visualizing stock prices.
    resources:
    - name: sql-warehouse
      sql_warehouse:
        id: warehouse_id
        permission: CAN_USE
    access_controls:
        - group_name: account users
          permission_level: CAN_USE
    '''
    app = models.resources.databricks.App.model_validate_yaml(io.StringIO(job_yaml))
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    budget_policy_id: str = Field(
        None, description="The Budget Policy ID set for this resource."
    )
    compute_size: str = Field(
        None,
        description="A string specifying compute size for the App. Possible values are `MEDIUM`, `LARGE`.",
    )

    description: str = Field(None, description="The description of the app.")

    git_repository: AppGitRepository = Field(None, description="")
    name: str = Field(
        None,
        description="The name of the app. The name must contain only lowercase alphanumeric characters and hyphens. It must be unique within the workspace.",
    )
    name_prefix: str = Field(None, description="Prefix added to the app name")
    name_suffix: str = Field(None, description="Suffix added to the app name")

    no_compute: bool = Field(None, description="")
    # provider_config: AppProviderConfig = Field(None, description="")
    resources: list[AppResource] = Field(
        None, description="A list of resources that the app have access to."
    )

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.name = self.name_prefix + self.name
                self.name_prefix = ""
            if self.name_suffix:
                self.name = self.name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    app_name=f"${{resources.{self.resource_name}.name}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:App"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "access_controls",
            "name_prefix",
            "name_suffix",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "resources": "resources",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_app"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal

from pydantic import Field
from pydantic import field_validator

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.models.datasources.customreader import CustomReader
from laktory.typing import AnyFrame

logger = get_logger(__name__)


class CustomDataSource(BaseDataSource):
    """
    Data source backed by a fully user-supplied read function. Use this when
    no built-in source type fits — REST APIs, proprietary databases, custom
    formats, etc.

    Laktory calls the function and wraps the returned DataFrame in Narwhals.
    Post-read operations (filter, selects, renames, drops) are still applied
    by Laktory after the custom read.

    Examples
    --------
    ```py
    from laktory import models

    source = models.CustomDataSource(
        custom_reader={
            "func_name": "mypackage.etl.my_read",
            "func_kwargs": {"table": "catalog.schema.my_table"},
        },
    )
    # df = source.read()
    ```

    ```py
    # mypackage/etl.py
    from laktory.models import LaktoryContext


    def my_read(table=None, laktory_context: LaktoryContext = None):
        import polars as pl

        return pl.scan_delta(table)
    ```

    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    type: Literal["CUSTOM"] = Field("CUSTOM", frozen=True, description="Source type")
    custom_reader: CustomReader = Field(
        ...,
        description=(
            "Custom reader definition. Can be set as a plain string (func_name only) "
            "or a full CustomReader object with func_name, func_args, and func_kwargs."
        ),
    )

    @field_validator("custom_reader", mode="before")
    @classmethod
    def coerce_custom_reader(cls, v):
        if isinstance(v, str):
            return {"func_name": v}
        return v

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return ["custom_reader"]

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return self.custom_reader.func_name

    # ----------------------------------------------------------------------- #
    # Read                                                                    #
    # ----------------------------------------------------------------------- #

    def _read(self, **kwargs) -> AnyFrame:
        return self.custom_reader.execute()



import inspect
from typing import TYPE_CHECKING
from typing import Any

from pydantic import Field
from pydantic import SkipValidation

from laktory.models.basemodel import BaseModel

if TYPE_CHECKING:
    pass


class LaktoryContext(BaseModel):
    """
    Runtime context object optionally injected into user-supplied functions by Laktory.

    Passed as the `laktory_context` keyword argument when the called function's
    signature explicitly declares it. Contains references to the
    pipeline objects active at call time.

    To opt in, add `laktory_context=None` to your function signature:

    ```python
    from laktory.models import LaktoryContext


    def my_write(df, laktory_context: LaktoryContext = None) -> None:
        print(laktory_context.sink)
        print(laktory_context.node)

        return df
    ```
    """

    node: SkipValidation[Any] = Field(
        None, description="Parent PipelineNode, or None when called outside a pipeline."
    )  # PipelineNode
    pipeline: SkipValidation[Any] = Field(
        None, description="Parent Pipeline, or None when called outside a pipeline."
    )  # Pipeline
    sink: SkipValidation[Any] = Field(
        None, description="Current data sink."
    )  # BaseDataSink  (set by CustomWriter)
    source: SkipValidation[Any] = Field(
        None, description="Current data source."
    )  # BaseDataSource (set by CustomReader)


def _build_laktory_context_kwargs(func, context: LaktoryContext) -> dict:
    """
    Return ``{'laktory_context': context}`` if *func* has an explicit
    ``laktory_context`` parameter, otherwise return ``{}``.

    Deliberately does not inject on ``**kwargs``-only signatures: many
    DataFrame methods (e.g. ``df.select``) accept ``**kwargs`` as named
    column expressions and would misinterpret a ``LaktoryContext`` value.
    """
    try:
        params = inspect.signature(func).parameters
    except (ValueError, TypeError):
        return {}
    return {"laktory_context": context} if "laktory_context" in params else {}



import importlib
from typing import Any

import narwhals as nw
from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.laktorycontext import LaktoryContext
from laktory.models.laktorycontext import _build_laktory_context_kwargs
from laktory.models.pipelinechild import PipelineChild

logger = get_logger(__name__)


class CustomWriter(BaseModel, PipelineChild):
    """
    Definition of a custom write function to be called when writing a DataFrame
    to a sink. Gives the user full control over how data is written.

    Laktory manages the streaming query lifecycle (foreachBatch wrapping, trigger,
    checkpoint, start/await) while the user-supplied function handles the actual
    write logic.

    The function is called as:

    ```python
    # func(df, *func_args, **func_kwargs)
    ```

    where `df` is the DataFrame (native or Narwhals depending on the sink's
    `dataframe_api` setting). Laktory optionally injects a `laktory_context`
    keyword argument containing pipeline runtime objects — declare it in your
    function signature to opt in.

    Examples
    --------
    ```py
    from laktory import models

    sink = models.FileDataSink(
        path="./my_table/",
        format="DELTA",
        custom_writer={
            "func_name": "mypackage.etl.my_write",
            "func_kwargs": {"extra_tag": "production"},
        },
    )
    # sink.write(df)
    ```

    ```py
    # mypackage/etl.py
    from laktory.models import LaktoryContext


    def my_write(
        df, extra_tag="default", laktory_context: LaktoryContext = None
    ) -> None:
        sink = laktory_context.sink  # access sink coordinates without re-hardcoding
        df.write.format("delta").mode("append").save(sink.path)
    ```
    """

    func_args: list[Any] = Field(
        [],
        description="Positional arguments passed to the function after the DataFrame.",
    )
    func_kwargs: dict[str, Any] = Field(
        {},
        description="Keyword arguments passed to the function.",
    )
    func_name: str = Field(
        ...,
        description=(
            "Fully qualified importable module path to the callable "
            "(e.g. 'mypackage.etl.my_write'). "
            "The function is imported at runtime via `importlib` and called with the "
            "DataFrame as its first argument. "
        ),
    )

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _resolve_func(self):
        """Dynamically import and return the callable from its dotted path."""
        module_path, func_name = self.func_name.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, func_name)

    def execute(self, df, context: LaktoryContext = None) -> None:
        """
        Invoke the user-supplied function with the DataFrame, configured
        args/kwargs, and optionally a `laktory_context` object.

        Parameters
        ----------
        df:
            Input DataFrame (native or Narwhals, as prepared by the caller).
        context:
            Pre-built LaktoryContext to inject. When ``None`` (the default),
            the context is built from the ``_parent`` chain at call time.
            Callers that run inside a ``foreachBatch`` lambda should pre-build
            the context before the lambda and pass it here so that the
            ``_parent`` references are captured while they are still intact.
        """
        func = self._resolve_func()

        if not isinstance(df, (nw.DataFrame, nw.LazyFrame)):
            df = nw.from_native(df)

        if self.dataframe_api == "NATIVE":
            df = nw.to_native(df)

        func_log = f"{self.func_name}("
        func_log += ",".join([str(a) for a in self.func_args])
        func_log += ",".join([f"{k}={v}" for k, v in self.func_kwargs.items()])
        func_log += f") with df type {type(df)}"
        logger.info(f"Writing df with {func_log}")

        if context is None:
            context = LaktoryContext(
                node=self.parent_pipeline_node,
                pipeline=self.parent_pipeline,
                sink=self.parent,
            )
        call_kwargs = {
            **self.func_kwargs,
            **_build_laktory_context_kwargs(func, context),
        }
        func(df, *self.func_args, **call_kwargs)



from .basedatasource import BaseDataSource
from .customdatasource import CustomDataSource
from .customreader import CustomReader
from .dataframedatasource import DataFrameDataSource
from .filedatasource import FileDataSource
from .hivemetastoredatasource import HiveMetastoreDataSource
from .pipelinenodedatasource import PipelineNodeDataSource
from .tabledatasource import TableDataSource
from .unitycatalogdatasource import UnityCatalogDataSource

classes = [
    CustomDataSource,
    FileDataSource,
    UnityCatalogDataSource,
    HiveMetastoreDataSource,
    DataFrameDataSource,
    PipelineNodeDataSource,
]

DataSourcesUnion = (
    CustomDataSource
    | FileDataSource
    | UnityCatalogDataSource
    | HiveMetastoreDataSource
    | DataFrameDataSource
    | PipelineNodeDataSource
)



from typing import TYPE_CHECKING

import narwhals as nw
from pydantic_core import CoreSchema
from pydantic_core import core_schema

if TYPE_CHECKING:
    pass

AnyFrame = nw.LazyFrame | nw.DataFrame


class VariableType(str):
    """
    Laktory variable ${vars.my_variable_name} or expression ${{ vars.id == 2}} string

    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/variables/)
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: callable
    ) -> CoreSchema:
        return core_schema.str_schema()



from .accesscontrol import AccessControl
from .accesscontrolruleset import AccessControlRuleSet
from .alert import Alert
from .app import App
from .catalog import Catalog
from .cluster import Cluster
from .clusterpolicy import ClusterPolicy
from .currentuser import CurrentUser
from .dashboard import Dashboard
from .dbfsfile import DbfsFile
from .directory import Directory
from .externallocation import ExternalLocation
from .grant import Grant
from .grants import Grants
from .group import Group
from .groupmember import GroupMember
from .job import Job
from .metastore import Metastore
from .metastoreassignment import MetastoreAssignment
from .metastoredataaccess import MetastoreDataAccess
from .mlflowexperiment import MLflowExperiment
from .mlflowmodel import MLflowModel
from .mlflowwebhook import MLflowWebhook
from .mwsnccbinding import MwsNccBinding
from .mwsnetworkconnectivityconfig import MwsNetworkConnectivityConfig
from .mwspermissionassignment import MwsPermissionAssignment
from .notebook import Notebook
from .notificationdestination import NotificationDestination
from .obotoken import OboToken
from .permissions import Permissions
from .pipeline import Pipeline
from .pythonpackage import PythonPackage
from .qualitymonitor import QualityMonitor
from .query import Query
from .recipient import Recipient
from .repo import Repo
from .schema import Schema
from .secret import Secret
from .secretacl import SecretAcl
from .secretscope import SecretScope
from .serviceprincipal import ServicePrincipal
from .serviceprincipalrole import ServicePrincipalRole
from .share import Share
from .storagecredential import StorageCredential
from .table import Table
from .user import User
from .userrole import UserRole
from .vectorsearchendpoint import VectorSearchEndpoint
from .vectorsearchindex import VectorSearchIndex
from .volume import Volume
from .warehouse import Warehouse
from .workspacebinding import WorkspaceBinding
from .workspacefile import WorkspaceFile
from .workspacetree import WorkspaceTree



from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class UserRole(BaseModel, PulumiResource, TerraformResource):
    role: str = Field(
        None, description="This is the id of the role or instance profile resource."
    )
    user_id: str = Field(None, description="This is the id of the user resource.")

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return f"{self.role}-{self.user_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:UserRole"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_user_role"



import importlib
from typing import Any

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.laktorycontext import LaktoryContext
from laktory.models.laktorycontext import _build_laktory_context_kwargs
from laktory.models.pipelinechild import PipelineChild

logger = get_logger(__name__)


class CustomReader(BaseModel, PipelineChild):
    """
    Definition of a custom read function used by `CustomDataSource`. Gives the
    user full control over how data is read.

    The function must return a DataFrame (native or Narwhals). Laktory optionally injects
    a `laktory_context` keyword argument — declare it in your function signature
    to opt in:

    ```python
    def my_read(laktory_context=None):
        print(laktory_context.source)
        print(laktory_context.node)
    ```

    Examples
    --------
    ```py
    from laktory import models

    source = models.CustomDataSource(
        custom_reader={
            "func_name": "mypackage.etl.my_read",
            "func_kwargs": {"table": "catalog.schema.my_table"},
        },
    )
    # df = source.read()
    ```

    ```py
    # mypackage/etl.py
    from laktory.models import LaktoryContext


    def my_read(table=None, laktory_context: LaktoryContext = None):
        import polars as pl

        return pl.scan_delta(table)
    ```
    """

    func_args: list[Any] = Field(
        [],
        description="Positional arguments passed to the function.",
    )
    func_kwargs: dict[str, Any] = Field(
        {},
        description="Keyword arguments passed to the function.",
    )
    func_name: str = Field(
        ...,
        description=(
            "Fully qualified importable module path to the callable "
            "(e.g. 'mypackage.etl.my_read'). "
            "The function is imported at runtime via `importlib` and must return "
            "a DataFrame (native or Narwhals)."
        ),
    )

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _resolve_func(self):
        """Dynamically import and return the callable from its dotted path."""
        module_path, func_name = self.func_name.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, func_name)

    def execute(self):
        """
        Invoke the user-supplied function with configured args/kwargs and
        optionally a `laktory_context` object.

        Returns
        -------
        :
            DataFrame returned by the user function (native or Narwhals).
        """
        func = self._resolve_func()

        func_log = f"{self.func_name}("
        func_log += ",".join([str(a) for a in self.func_args])
        func_log += ",".join([f"{k}={v}" for k, v in self.func_kwargs.items()])
        func_log += ")"
        logger.info(f"Reading with {func_log}")

        context = LaktoryContext(
            node=self.parent_pipeline_node,
            pipeline=self.parent_pipeline,
            source=self.parent,
        )
        call_kwargs = {
            **self.func_kwargs,
            **_build_laktory_context_kwargs(func, context),
        }
        return func(*self.func_args, **call_kwargs)



from collections import defaultdict
from typing import Any

import networkx as nx
from pydantic import Field
from pydantic import SkipValidation
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline._execute import _execute  # noqa: F401
from laktory.models.pipeline._post_execute import _post_execute  # noqa: F401
from laktory.models.pipeline.pipeline import Pipeline
from laktory.models.pipeline.pipelinetask import PipelineTask

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Functions                                                            #
# --------------------------------------------------------------------------- #


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class PipelineExecutionPlan(BaseModel):
    """
    A pipeline execution plan defines the pipeline tasks to be executed according to the selected nodes/tasks/tags
    and their dependencies. It constructs a DAG of pipeline tasks, where each task can consist of one or more nodes
    that share the same `execution_task_name`.
    """

    pipeline: SkipValidation[Pipeline] = Field(
        ...,
        description="""Pipeline""",
    )
    selects: list[str] | None = Field(
        None,
        description="""
        List of node names with optional dependency notation:
        
        - `{node_name}`: Execute the node only.
        - `*{node_name}`: Execute the node and its upstream dependencies.
        - `{node_name}*`: Execute the node and its downstream dependencies.
        - `*{node_name}*`: Execute the node, its upstream, and downstream dependencies.
        """,
    )

    @model_validator(mode="after")
    def update_pipeline(self) -> Any:
        if self.pipeline is not None:
            self.pipeline._plan = self
        return self

    # -------------------------------------------------------------------------------- #
    # Nodes                                                                            #
    # -------------------------------------------------------------------------------- #

    @property
    def node_names(self) -> list[str]:
        """Selected pipeline node names"""
        # TODO: Refactor as a cached property??
        if self.selects is None:
            return self.pipeline.sorted_node_names

        # Validate and resolve selected nodes/tasks/tags
        selected_nodes = set()
        dag = self.pipeline.dag

        for item in self.selects:
            if item.startswith("*") and item.endswith("*"):
                # Node + both upstream and downstream dependencies
                node_name = item[1:-1]
                if node_name in self.pipeline.nodes_dict:
                    selected_nodes.add(node_name)
                    selected_nodes.update(nx.ancestors(dag, node_name))
                    selected_nodes.update(nx.descendants(dag, node_name))
                else:
                    # Check if it's a task or tag
                    task_or_tag_nodes = [
                        n.name
                        for n in self.pipeline.nodes
                        if n.execution_task_name == node_name or node_name in n.tags
                    ]
                    if not task_or_tag_nodes:
                        raise ValueError(
                            f"Invalid node, execution_task_name, or tag: {node_name}"
                        )
                    for group_node in task_or_tag_nodes:
                        selected_nodes.add(group_node)
                        selected_nodes.update(nx.ancestors(dag, group_node))
                        selected_nodes.update(nx.descendants(dag, group_node))
            elif item.startswith("*"):
                # Node + upstream dependencies
                node_name = item[1:]
                if node_name in self.pipeline.nodes_dict:
                    selected_nodes.add(node_name)
                    selected_nodes.update(nx.ancestors(dag, node_name))
                else:
                    # Check if it's a task or tag
                    task_or_tag_nodes = [
                        n.name
                        for n in self.pipeline.nodes
                        if n.execution_task_name == node_name or node_name in n.tags
                    ]
                    if not task_or_tag_nodes:
                        raise ValueError(
                            f"Invalid node, execution_task_name, or tag: {node_name}"
                        )
                    for group_node in task_or_tag_nodes:
                        selected_nodes.add(group_node)
                        selected_nodes.update(nx.ancestors(dag, group_node))
            elif item.endswith("*"):
                # Node + downstream dependencies
                node_name = item[:-1]
                if node_name in self.pipeline.nodes_dict:
                    selected_nodes.add(node_name)
                    selected_nodes.update(nx.descendants(dag, node_name))
                else:
                    # Check if it's a task or tag
                    task_or_tag_nodes = [
                        n.name
                        for n in self.pipeline.nodes
                        if n.execution_task_name == node_name or node_name in n.tags
                    ]
                    if not task_or_tag_nodes:
                        raise ValueError(
                            f"Invalid node, execution_task_name, or tag: {node_name}"
                        )
                    for group_node in task_or_tag_nodes:
                        selected_nodes.add(group_node)
                        selected_nodes.update(nx.descendants(dag, group_node))
            else:
                # Node only
                if item in self.pipeline.nodes_dict:
                    selected_nodes.add(item)
                else:
                    # Check if it's a task or tag
                    task_or_tag_nodes = [
                        n.name
                        for n in self.pipeline.nodes
                        if n.execution_task_name == item or item in n.tags
                    ]
                    if not task_or_tag_nodes:
                        raise ValueError(
                            f"Invalid node, execution_task_name, or tag: {item}"
                        )
                    selected_nodes.update(task_or_tag_nodes)

        # Sort nodes based on DAG topological order
        sorted_node_names = [
            node for node in self.pipeline.sorted_node_names if node in selected_nodes
        ]

        return sorted_node_names

    @property
    def nodes_dict(self):
        """Selected pipeline nodes dict"""
        return {k: self.pipeline.nodes_dict[k] for k in self.node_names}

    @property
    def nodes_dag(self) -> nx.DiGraph:
        """Selected pipeline nodes DAG"""
        selected_nodes = set(self.node_names)
        dag = self.pipeline.dag
        for node_name in list(dag.nodes.keys()):
            if node_name not in selected_nodes:
                dag.remove_node(node_name)

        return dag

    # -------------------------------------------------------------------------------- #
    # Tasks                                                                            #
    # -------------------------------------------------------------------------------- #

    @property
    def dag(self) -> nx.DiGraph:
        """
        Constructs a DAG for the execution plan, where each node is either:
        - A single pipeline node (default `execute_task_name`), or
        - A group of nodes (if they share the same `execute_task_name`).

        Pipeline nodes without sink(s) are omitted unless they are leaf (terminal) nodes
        as they will be executed implicitly when executing their downstream nodes.

        Returns
        -------
        nx.DiGraph
            A directed acyclic graph representing the execution plan.
        """
        nodes_dag = self.nodes_dag
        nodes_dict = self.nodes_dict

        def is_valid(node):
            return node.has_sinks or nodes_dag.out_degree[node.name] == 0

        # Create a mapping of execution nodes (groups or standalone) to their pipeline nodes
        task_nodes = defaultdict(list)

        for node in nodes_dict.values():
            if not is_valid(node):
                continue

            task_name = f"{node.execution_task_name}"
            task_nodes[task_name].append(node.name)

        # Initialize a new DAG
        dag = nx.DiGraph()

        # Add execution nodes
        for task_name, node_names in task_nodes.items():
            dag.add_node(task_name, node_names=node_names)

        # Map each included pipeline node to its task name
        nodes_task_name = {}
        for task_name, node_names in task_nodes.items():
            for node_name in node_names:
                nodes_task_name[node_name] = task_name

        def resolve_upstream_task_names(node_name: str) -> list[str]:
            """
            Walk upstream from start_name, skipping selected nodes without sinks,
            until we reach selected nodes with sinks (i.e., execution nodes).
            """
            resolved = set()
            stack = list(nodes_dag.predecessors(node_name))
            seen = set()

            while stack:
                pred_name = stack.pop()
                if pred_name in seen:
                    continue
                seen.add(pred_name)

                pred_node = nodes_dict.get(pred_name)
                if pred_node is None:
                    # Shouldn't happen if nodes_dag is consistent, but be safe.
                    continue

                if is_valid(pred_node):
                    # This upstream node is a task node (group or standalone)
                    task_name = nodes_task_name.get(pred_name)
                    if task_name is not None:
                        resolved.add(task_name)
                    continue

                # Not a task node: skip through it (it will run implicitly downstream)
                stack.extend(list(nodes_dag.predecessors(pred_name)))

            return list(resolved)

        # Add edges
        for task_name, node_names in task_nodes.items():
            upstream_task_names = []
            for node_name in node_names:
                upstream_task_names += resolve_upstream_task_names(node_name)
            upstream_task_names = list(set(upstream_task_names))

            for upstream_task_name in upstream_task_names:
                if upstream_task_name != task_name:
                    dag.add_edge(upstream_task_name, task_name)

        return dag

    @property
    def tasks(self):
        """
        List of pipeline tasks to be executed. Each task is either:
        - A single pipeline node (default `execute_task_name`), or
        - A group of nodes (if they share the same `execute_task_name`).

        Pipeline nodes without sink(s) are omitted unless they are leaf (terminal) nodes
        as they will be executed implicitly when executing their downstream nodes.

        Returns
        -------
        output: list[PipelineTask]
        """
        tasks = []
        for dag_node, attr in self.dag.nodes(data=True):
            tasks += [
                PipelineTask(
                    name=dag_node,
                    pipeline=self.pipeline,
                    node_names=attr["node_names"],
                )
            ]

        return tasks

    @property
    def tasks_dict(self):
        """Tasks dictionary"""
        return {task.name: task for task in self.tasks}



"""Custom Griffe extension: sort Pydantic fields and hide infrastructure base class members."""

from __future__ import annotations

from typing import Any

from griffe import Attribute
from griffe import Class
from griffe import DocstringSectionAttributes
from griffe import DocstringSectionParameters
from griffe import Expr
from griffe import ExprBinOp
from griffe import ExprName
from griffe import ExprSubscript
from griffe import Extension
from griffe import dynamic_import
from griffe import get_logger

logger = get_logger("laktory.griffe_ext")


def _replace_expr_name(
    expr: str | Expr | None,
    name: str,
    replacement: ExprName,
) -> str | Expr | None:
    """Recursively replace ExprName(name) nodes inside an expression tree."""
    if expr is None or isinstance(expr, str):
        return expr
    if isinstance(expr, ExprName):
        return replacement if expr.name == name else expr
    if isinstance(expr, ExprBinOp):
        expr.left = _replace_expr_name(expr.left, name, replacement)
        expr.right = _replace_expr_name(expr.right, name, replacement)
    elif isinstance(expr, ExprSubscript):
        expr.left = _replace_expr_name(expr.left, name, replacement)
        expr.slice = _replace_expr_name(expr.slice, name, replacement)
    return expr


class FieldSorterExtension(Extension):
    """Sort Pydantic fields alphabetically and hide infrastructure base class members.

    Base classes marked with ``__doc_hide_base__ = True`` have ALL their fields
    and public methods hidden in child class documentation. All remaining fields
    (own + domain-parent) are merged into one sorted "PARAMETER" section.

    Individual child classes can additionally hide specific inherited fields::

        class MyModel(ParentModel):
            __hidden_doc_fields__: ClassVar[set[str]] = {"some_field"}
    """

    def on_class(self, *, cls: Class, **kwargs: Any) -> None:
        if not cls.docstring:
            return

        try:
            runtime_obj = dynamic_import(cls.path)
        except ImportError:
            logger.debug("Could not import %s", cls.path)
            return

        # Identify which ancestor classes are "hidden bases"
        hidden_base_names: set[str] = {
            ancestor.__name__
            for ancestor in runtime_obj.__mro__[1:]
            if ancestor.__dict__.get("__doc_hide_base__", False)
        }

        # Collect field names owned by hidden base classes
        hidden_field_names: set[str] = set()
        for ancestor in runtime_obj.__mro__[1:]:
            if ancestor.__name__ in hidden_base_names:
                hidden_field_names |= set(
                    ancestor.__dict__.get("__annotations__", {}).keys()
                )

        # Collect per-class explicitly hidden fields (__hidden_doc_fields__ accumulates across MRO)
        per_class_hidden: set[str] = set()
        for ancestor in runtime_obj.__mro__:
            per_class_hidden |= set(
                ancestor.__dict__.get("__hidden_doc_fields__", set())
            )

        all_hidden_fields = hidden_field_names | per_class_hidden

        # Shadow public members of hidden griffe base classes with no-docstring stubs.
        # Because inherited_members only includes names NOT in cls.members, the stub
        # prevents the real inherited member from appearing. show_if_no_docstring=false
        # then hides the stub itself.
        if hidden_base_names:
            try:
                for base_griffe_cls in cls.mro():
                    if base_griffe_cls.name in hidden_base_names:
                        for member_name in list(base_griffe_cls.members.keys()):
                            if (
                                not member_name.startswith("_")
                                and member_name not in cls.members
                            ):
                                cls.members[member_name] = Attribute(
                                    member_name, parent=cls
                                )
            except (ValueError, Exception) as e:
                logger.debug("MRO walk failed for %s: %s", cls.path, e)

        # Filter hidden fields and sort all remaining fields alphabetically
        sections = cls.docstring.parsed
        for section in list(sections):
            if not isinstance(
                section, (DocstringSectionParameters, DocstringSectionAttributes)
            ):
                continue
            visible = [
                item for item in section.value if item.name not in all_hidden_fields
            ]
            visible.sort(key=lambda x: x.name)
            section.value = visible

        # Qualify bare VariableType ExprName nodes so autorefs can link them.
        # griffe_fieldz uses display_as_type() which strips the module prefix,
        # producing ExprName('VariableType') with no parent context.  We fix
        # this by setting parent to the laktory.typing griffe module so that
        # canonical_path resolves to 'laktory.typing.VariableType'.
        try:
            typing_module = cls.package["typing"]
        except (KeyError, AttributeError):
            typing_module = None

        if typing_module is not None:
            qualified = ExprName("VariableType", parent=typing_module)
            for section in cls.docstring.parsed:
                if not isinstance(
                    section, (DocstringSectionParameters, DocstringSectionAttributes)
                ):
                    continue
                for item in section.value:
                    if item.annotation is not None:
                        item.annotation = _replace_expr_name(
                            item.annotation, "VariableType", qualified
                        )

        # Remove empty parameter/attribute sections
        cls.docstring.parsed = [
            s
            for s in sections
            if not isinstance(
                s, (DocstringSectionParameters, DocstringSectionAttributes)
            )
            or s.value
        ]



from pydantic import Field
from pydantic import SkipValidation

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline._execute import _execute  # noqa: F401
from laktory.models.pipeline._post_execute import _post_execute  # noqa: F401
from laktory.models.pipeline.pipeline import Pipeline
from laktory.typing import AnyFrame

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Functions                                                            #
# --------------------------------------------------------------------------- #


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class PipelineTask(BaseModel):
    """
    A pipeline task is a unit of execution within a pipeline, defined by a set of nodes to be executed together.
    """

    name: str = Field(
        ...,
        description="""Pipeline task name""",
    )
    pipeline: SkipValidation[Pipeline] = Field(
        ...,
        description="""Pipeline""",
    )
    node_names: list[str] = Field(
        ..., description="""List of node names in sorted order of execution"""
    )

    def execute(
        self,
        write_sinks=True,
        full_refresh: bool = False,
        named_dfs: dict[str, AnyFrame] = None,
        update_tables_metadata: bool = True,
    ) -> None:
        """
        Execute the pipeline task.

        Parameters
        ----------
        write_sinks:
            If `False` writing of node sinks will be skipped
        full_refresh:
            If `True` all nodes will be completely re-processed by deleting
            existing data and checkpoints before processing.
        named_dfs:
            Named DataFrames to be passed to pipeline nodes transformer.
        update_tables_metadata:
            Update tables metadata
        """

        logger.info(f"Executing pipeline task '{self.name}'")

        # Execute nodes
        for node_name in self.node_names:
            node = self.pipeline.nodes_dict[node_name]
            if named_dfs is None:
                named_dfs = {}

            node.execute(
                write_sinks=write_sinks,
                full_refresh=full_refresh,
                named_dfs=named_dfs,
                update_tables_metadata=update_tables_metadata,
            )

    @property
    def upstream_task_names(self) -> list[str]:
        """Get upstream task names"""
        plan = self.pipeline._plan
        names = []
        for edges in plan.dag.in_edges(self.name):
            if edges[0] != self.name:
                names += [edges[0]]
        return names

    @property
    def nodes(self):
        """Task nodes"""
        return [self.pipeline.nodes_dict[node_name] for node_name in self.node_names]

    @property
    def has_sinks(self) -> bool:
        """`True` if at least one sink is found in task nodes."""
        has_sinks = False
        for node in self.nodes:
            if node.has_sinks:
                has_sinks = True
                break
        return has_sinks



from .orchestrators.airfloworchestrator import AirflowOrchestrator
from .orchestrators.databricksjoborchestrator import DatabricksJobOrchestrator
from .orchestrators.databrickspipelineorchestrator import DatabricksPipelineOrchestrator
from .pipeline import Pipeline
from .pipeline import PipelineNode
from .pipelineexecutionplan import PipelineExecutionPlan
from .pipelinetask import PipelineTask



import re
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Any
from typing import Literal

from pydantic import Field
from pydantic import field_validator

from laktory.models.basemodel import BaseModel
from laktory.models.pipelinechild import PipelineChild


def _str_to_timedelta(v):
    pattern = re.compile(r"(\d+)([smhd])")

    units = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days"}
    kwargs = {}

    for value, unit in pattern.findall(v):
        kwargs[units[unit]] = kwargs.get(units[unit], 0) + int(value)

    return timedelta(**kwargs)


class CronSchedule(BaseModel):
    """
    Cron Schedule.
    """

    cron: str = Field(
        ...,
        description="Cron expression defining when the DAG should be scheduled to run.",
    )
    timezone: str = Field(
        "utc", description="Timezone used to interpret the cron schedule."
    )


class AirflowOrchestrator(PipelineChild):
    """
    Airflow used as an orchestrator to execute a Laktory pipeline.

    Examples
    --------
    ```python
    import laktory as lk

    o = lk.models.AirflowOrchestrator(
        description="A Laktory pipeline",
        schedule={
            "cron": "0 0 * * *",
            "timezone": "utc",
        },
        start_date="2026-03-03",
        max_active_runs=1,
    )
    print(o)
    '''
    dataframe_backend_=None dataframe_api_=None type='AIRFLOW' description='A Laktory pipeline' schedule=CronSchedule(variables={}, cron='0 0 * * *', timezone='utc') start_date=datetime.datetime(2026, 3, 3, 0, 0, tzinfo=datetime.timezone.utc) end_date=None template_searchpath=None user_defined_macros=None user_defined_filters=None default_args=None max_active_tasks=None max_active_runs=1 max_consecutive_failed_dag_runs=None dagrun_timeout=None catchup=None doc_md=None access_control=None is_paused_upon_creation=None jinja_environment_kwargs=None render_template_as_native_obj=None tags=None owner_links=None auto_register=None fail_fast=None dag_display_name=None disable_bundle_versioning=None dataframe_backend=<DataFrameBackends.PYSPARK: 'PYSPARK'> dataframe_api='NARWHALS'
    '''
    ```

    References
    ----------
    * [Data Pipeline](https://www.laktory.ai/concepts/pipeline/)
    """

    type: Literal["AIRFLOW"] = Field("AIRFLOW", description="Type of orchestrator")
    description: str = Field(
        None, description="Optional human-readable description of the DAG."
    )
    schedule: CronSchedule = Field(
        None,
        description="Schedule that defines when the DAG should run. Uses a cron expression and timezone.",
    )  # simplified from airflow schedule options
    start_date: datetime = Field(
        None,
        description="The earliest date from which the scheduler will start creating DAG runs.",
    )
    end_date: datetime = Field(
        None, description="Optional date after which no new DAG runs will be scheduled."
    )
    template_searchpath: str | list[str] = Field(
        None,
        description="Directories to search for template files referenced in templated fields.",
    )
    # template_undefined: Type[jinja2.StrictUndefined] = jinja2.StrictUndefined  # not serializable
    user_defined_macros: dict = Field(
        None,
        description="Custom macros made available in the Jinja template context for all tasks in the DAG.",
    )
    user_defined_filters: dict = Field(
        None,
        description="Custom Jinja filters available when rendering templated fields.",
    )
    default_args: dict[str, Any] = Field(
        None,
        description="Dictionary of default arguments applied to all tasks in the DAG unless overridden.",
    )
    max_active_tasks: int = Field(
        None,
        description="Maximum number of task instances allowed to run concurrently across all DAG runs.",
    )
    max_active_runs: int = Field(
        None, description="Maximum number of active DAG runs allowed at the same time."
    )
    max_consecutive_failed_dag_runs: int = Field(
        None,
        description="Number of consecutive failed DAG runs allowed before the DAG is automatically paused.",
    )
    dagrun_timeout: str | timedelta = Field(
        None,
        description="Maximum time allowed for a DAG run to complete before it is marked as failed.",
    )
    catchup: bool = Field(
        None,
        description="Whether the scheduler should create DAG runs for all past schedule intervals between the start date and the current time.",
    )
    # on_success_callback: None | DagStateChangeCallback | list[DagStateChangeCallback] = None  # not serializable
    # on_failure_callback: None | DagStateChangeCallback | list[DagStateChangeCallback] = None  # not serializable
    # deadline: list[DeadlineAlert] | DeadlineAlert = None   # not serializable
    doc_md: str = Field(
        None,
        description="Markdown documentation displayed in the Airflow UI for the DAG.",
    )
    # params: ParamsDict | dict[str, Any] = None  # forced by Laktory
    access_control: dict[str, dict[str, list[str]]] | dict[str, list[str]] = Field(
        None,
        description="Access control configuration defining roles and permissions for this DAG.",
    )
    is_paused_upon_creation: bool = Field(
        None, description="Whether the DAG should be paused when first created."
    )
    jinja_environment_kwargs: dict = Field(
        None,
        description="Additional keyword arguments used when creating the Jinja template environment.",
    )
    render_template_as_native_obj: bool = Field(
        None,
        description="If true, Jinja templates will render native Python objects instead of strings.",
    )
    tags: list[str] = Field(
        None, description="Tags used to organize and filter DAGs in the Airflow UI."
    )
    owner_links: dict[str, str] = Field(
        None,
        description="Mapping of DAG owners to URLs that provide additional information about them.",
    )
    auto_register: bool = Field(
        None,
        description="Whether the DAG should automatically be registered when parsed by Airflow.",
    )
    fail_fast: bool = Field(
        None,
        description="If enabled, the DAG run will stop scheduling new tasks as soon as a task fails.",
    )
    dag_display_name: str = Field(
        None,
        description="Human-friendly name displayed in the Airflow UI instead of the DAG ID.",
    )
    disable_bundle_versioning: bool = Field(
        None,
        description="If enabled, disables versioning of DAG bundles when deploying.",
    )

    @field_validator("dagrun_timeout", mode="before")
    def validate_timedelta(cls, v: Any) -> Any:
        return _str_to_timedelta(v)

    @field_validator("start_date", "end_date", mode="before")
    def validate_datetime(cls, v: Any) -> Any:
        if isinstance(v, str):
            v = datetime.fromisoformat(v)
        if isinstance(v, datetime):
            if v.tzinfo is None:
                v = v.replace(tzinfo=timezone.utc)
        return v

    @property
    def _schedule(self):
        schedule = self.schedule
        if isinstance(self.schedule, CronSchedule):
            from airflow.timetables.interval import CronDataIntervalTimetable

            schedule = CronDataIntervalTimetable(
                cron=schedule.cron,
                timezone=schedule.timezone,
            )

        return schedule

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def to_airflow(self, **dag_kwargs):
        from airflow.sdk import dag
        from airflow.sdk import get_current_context
        from airflow.sdk import task

        pl = self.parent_pipeline
        plan = pl.get_execution_plan()

        def register_task(pl_task):
            @task(task_id=pl_task.name)
            def airflow_task():
                ctx = get_current_context()

                # Values from dag_run.conf (Trigger DAG JSON)
                conf = (ctx.get("dag_run").conf or {}) if ctx.get("dag_run") else {}

                # Defaults from DAG params (if set)
                params = ctx.get("params") or {}

                # Let conf override params override hard defaults
                full_refresh = conf.get(
                    "full_refresh", params.get("full_refresh", False)
                )

                pl_task.execute(
                    full_refresh=bool(full_refresh),
                )

            return airflow_task

        kwargs = {
            "dag_id": pl.name,
            "params": {
                "full_refresh": False,
            },
        }
        for fname in self.model_fields_set:
            if fname in ["type"]:
                continue
            kwargs[fname] = getattr(self, fname)
            if fname == "schedule":
                kwargs[fname] = self._schedule
        for k, v in dag_kwargs.items():
            kwargs[k] = v

        @dag(**kwargs)
        def airflow_dag():
            tasks = {}
            nodes = {}

            # Create tasks and nodes
            for pl_task in plan.tasks:
                tasks[pl_task.name] = register_task(pl_task)
                nodes[pl_task.name] = tasks[pl_task.name]()

            # Set dependencies
            for task_name in tasks.keys():
                for upstream_task_name in plan.tasks_dict[
                    task_name
                ].upstream_task_names:
                    nodes[upstream_task_name] >> nodes[task_name]

        return airflow_dag()
